package com.emarcotte.glfoo.extras;

import java.nio.FloatBuffer;

import javax.media.opengl.GL;

import com.emarcotte.glfoo.GraphicsDriver;
import com.emarcotte.glfoo.Renderable;

/**
 * Edged Renderables draw (fairly effectively) the edges around some 3d Object
 * (not just the silhouette).
 * 
 * The technique is based on Jeff Lander's method described in "Under the Shade
 * of the Rendering Tree."
 * 
 * @author Eugene Marcotte
 */
public abstract class EdgedRenderable implements Renderable {

	/**
	 * The color of the edge to draw.
	 */
	private FloatBuffer color;

	private boolean lighting;

	/**
	 * Default constructor, simply set the color to black.
	 */
	public EdgedRenderable() {
		color = FloatBuffer.wrap(new float[] { 0, 0, 0, 1 });
	}

	/**
	 * The draw method invokes both calls to <code>drawObject</code> and sets
	 * up the edge rendering.
	 * 
	 * @param driver
	 *            The GraphicsDriver that is currently running.
	 */

	public final void draw(GraphicsDriver driver) {
		GL gl = driver.getGl();

		gl.glEnable(GL.GL_CULL_FACE);

		// use the front face to draw on
		gl.glPolygonMode(GL.GL_FRONT, GL.GL_FILL);
		gl.glDepthFunc(GL.GL_LESS);
		gl.glCullFace(GL.GL_BACK);

		drawObject(driver, true);

		// use the back faces with big lines to show the outline
		gl.glLineWidth(5.0f);
		gl.glPolygonMode(GL.GL_BACK, GL.GL_LINE);
		gl.glDepthFunc(GL.GL_LEQUAL);
		gl.glCullFace(GL.GL_FRONT);

		lighting = gl.glIsEnabled(GL.GL_LIGHTING);

		if (lighting) {
			gl.glDisable(GL.GL_LIGHTING);
		}

		gl.glColor4fv(color);
		drawObject(driver, false);

		if (lighting) {
			gl.glEnable(GL.GL_LIGHTING);
		}
		gl.glCullFace(GL.GL_BACK);
	}

	/**
	 * Set the color of the border.
	 * 
	 * @param r
	 *            Red component
	 * @param b
	 *            Blue component
	 * @param g
	 *            Green component
	 * @param a
	 *            The Alpha component
	 */
	public void setColor(float r, float g, float b, float a) {
		color.put(0, r);
		color.put(1, g);
		color.put(2, b);
		color.put(3, a);
	}

	/**
	 * Defined by the sub-class, drawObject performs the actual drawing of the
	 * object that will be edged. This method will be called twice per
	 * rendering. The first time it is invoked, the parameter <code>first</code>
	 * will be set to true allowing the sub-class's implementation to vary
	 * functionality for first and second renderings.
	 * 
	 * The first rendering is used to generate the basic object that will be
	 * displayed.
	 * 
	 * The second rendering is used to generate the edges of the object.
	 * 
	 * @param driver
	 *            The GraphicsDriver that is currently running.
	 * @param first
	 *            Specifies that this is the first or second rendering of the
	 *            object this cycle.
	 */
	protected abstract void drawObject(GraphicsDriver driver, boolean first);

}
