package com.emarcotte.glfoo.extras;

import com.emarcotte.glfoo.GraphicsDriver;
import com.emarcotte.glfoo.Renderable;
import com.emarcotte.glfoo.implementation.DefaultRenderer;

/**
 * Post processing renderer utilizes the before and after phases of rendering to
 * inject effect logic into the pipeline. This logic is in the form of a
 * {@link PostProcessingEffect}.
 * 
 * @author Eugene Marcotte <emarcotte@gmail.com>
 */
public class PostProcessingRenderer extends DefaultRenderer<Renderable> {

	/**
	 * The current effect. TODO: Should the effect be swappable during runtime?
	 */
	private PostProcessingEffect effect;

	private boolean effectInitialized;

	/**
	 * Default constructor.
	 * 
	 * @param e
	 *            The effect to use.
	 * 
	 * @throws IllegalArgumentException
	 *             When the effect is null.
	 */
	public PostProcessingRenderer(PostProcessingEffect e) {
		if (e == null) {
			throw new IllegalArgumentException("Effect is null");
		}
		effect = e;
		effectInitialized = false;
	}

	@Override
	public void initialize(GraphicsDriver driver) {
		if (!effectInitialized) {
			effectInitialized = true;
			effect.init(driver);
		}
	}

	/**
	 * The before phase of rendering. Calls init if needed and then calls into
	 * the effect's setupFrame method.
	 * 
	 * @param driver
	 *            The driver running currently.
	 */
	@Override
	public void renderBefore(GraphicsDriver driver) {
		effect.setupFrame(driver);
	}

	/**
	 * The render phase. Simply draws all screen objects.
	 * 
	 * @param driver
	 *            The driver that is running.
	 * @param r
	 *            The current renderable to draw.
	 */
	@Override
	public void render(GraphicsDriver driver, Renderable r) {
		r.draw(driver);
	}

	@Override
	public void renderAfter(GraphicsDriver driver) {
		effect.processFrame(driver);
	}
}
