package com.emarcotte.glfoo.implementation;

import java.awt.Frame;
import java.awt.event.KeyListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;

import javax.media.opengl.GLCanvas;
import javax.media.opengl.GLCapabilities;
import javax.media.opengl.GLEventListener;
import javax.swing.SwingUtilities;

import com.emarcotte.glfoo.MainWindow;
import com.emarcotte.glfoo.MouseHandler;
import com.sun.opengl.util.FPSAnimator;

/**
 * 
 * @author Eugene Marcotte
 */
public class DefaultMainWindow extends WindowAdapter implements MainWindow {
	private GLCapabilities caps;

	/**
	 * Window where things are displayed.
	 */
	private Frame window;

	/**
	 * Animator for the window.
	 */
	private FPSAnimator animator;

	/**
	 * The OpenGL canvas.
	 */
	private GLCanvas canvas;

	/**
	 * The event listener for the GLCanvas.
	 */
	private GLEventListener driver;

	public DefaultMainWindow(GLEventListener driver) {
		this.driver = driver;
	}

	private final class WindowCreator implements Runnable {
		public void run() {
			DefaultMainWindow.this.create();
		}
	}

	private final class WindowStarter implements Runnable {
		public void run() {
			DefaultMainWindow.this.start();
		}
	}

	/**
	 * Creates the GUI.
	 */
	@Override
	public void create() {
		if (SwingUtilities.isEventDispatchThread()) {
			caps = new GLCapabilities();
			caps.setDoubleBuffered(true);
			caps.setHardwareAccelerated(true);

			caps.setSampleBuffers(true);
			caps.setNumSamples(4);

			canvas = new GLCanvas(caps);
			canvas.addGLEventListener(driver);

			animator = new FPSAnimator(canvas, 60);

			window = new Frame("Modelviewer");
			window.add(canvas);

			window.setSize(640, 480);
			window.addWindowListener(DefaultMainWindow.this);

			ExceptionDialog exceptionDialog = new ExceptionDialog();
			exceptionDialog.setTitle("Graphics Exception");
			exceptionDialog.setDisposeAction(new Runnable() {
				public void run() {
					DefaultMainWindow.this.close();
				}
			});

			exceptionDialog.setLaunchAction(new Runnable() {
				public void run() {
					DefaultMainWindow.this.disable();
				}
			});

			Thread.currentThread().setUncaughtExceptionHandler(exceptionDialog);

		} else {
			try {
				SwingUtilities.invokeAndWait(new WindowCreator());
			} catch (Exception e) {
				throw new RuntimeException("Failed to create MainWindow", e);
			}
		}
	}

	@Override
	public void start() {
		if (SwingUtilities.isEventDispatchThread()) {
			window.setVisible(true);
			animator.start();
			canvas.requestFocus();
		} else {
			try {
				SwingUtilities.invokeAndWait(new WindowStarter());
			} catch (Exception e) {
				throw new RuntimeException("Failed to start MainWindow", e);
			}
		}
	}

	/**
	 * @param e
	 */
	public void windowClosing(WindowEvent e) {
		close();
	}

	public void close() {
		if (animator.isAnimating()) {
			animator.stop();
		}
		window.setVisible(false);
		window.dispose();
	}

	/**
	 * Adds a MouseHandler.
	 * 
	 * @param mouseHandler
	 *            The new MouseHandler.
	 */
	@Override
	public void addMouseHandler(MouseHandler mouseHandler) {
		canvas.addMouseListener(mouseHandler);
		canvas.addMouseMotionListener(mouseHandler);
	}

	/**
	 * Remove a MouseHandler.
	 * 
	 * @param mouseHandler
	 *            The handler to remove.
	 */
	@Override
	public void removeMouseHandler(MouseHandler mouseHandler) {
		canvas.removeMouseMotionListener(mouseHandler);
		canvas.removeMouseListener(mouseHandler);
	}

	/**
	 * Adds a KeyListener.
	 * 
	 * @param keyListener
	 *            The listener to add.
	 */
	@Override
	public void addKeyListener(KeyListener keyListener) {
		canvas.addKeyListener(keyListener);
	}

	/**
	 * Removes a KeyListener.
	 * 
	 * @param keyListener
	 *            The listener to remove.
	 */
	@Override
	public void removeKeyListener(KeyListener keyListener) {
		canvas.removeKeyListener(keyListener);
	}

	public void disable() {
		animator.stop();
		canvas.setEnabled(false);
		canvas.getParent().remove(canvas);
	}
}
