/**
 * 
 */
package com.emarcotte.glfoo.implementation;

import java.util.Collection;
import java.util.LinkedList;

import javax.media.opengl.GL;

import com.emarcotte.glfoo.Camera;
import com.emarcotte.glfoo.GraphicsDriver;
import com.emarcotte.glfoo.Renderable;
import com.emarcotte.glfoo.Renderer;

/**
 * Renderers provide the graphics driver with Renderables as well as a minimal
 * amount of opengl setup.
 * 
 * The camera usage in this renderer is quite basic for the most part it simply
 * applies the perspective and position. The camera implementation by default is
 * a DefaultCamera, but does not necessarily have to be.
 * 
 * @author Eugene Marcotte <emarcotte@gmail.com>
 *                                                                                                                                                     
 * @param <E> The Renderable type that this Renderer can handle.
 */
public abstract class DefaultRenderer<E extends Renderable> implements Renderer {
	/**
	 * The renderables to draw each frame.
	 */
	private Collection<E> renderables;

	/**
	 * The renderables that have been added to this renderer, have not been
	 * rendered yet, and are waiting to be initialized.
	 */
	private Collection<E> newRenderables;

	/**
	 * Simple 3d camera, no fancy stuff.
	 */
	private Camera camera;

	/**
	 * Sets up the renderer with a new {@link DefaultCamera}
	 */
	public DefaultRenderer() {
		this(new DefaultCamera());
	}

	/**
	 * Sets up the renderer with a provided camera
	 * 
	 * @param c
	 *            Some camera.
	 */
	public DefaultRenderer(Camera c) {
		renderables = new LinkedList<E>();
		newRenderables = new LinkedList<E>();
		setCamera(c);
	}

	public void setCamera(Camera c) {
		this.camera = c;
	}

	/**
	 * @see Renderer#getCamera()
	 */
	@Override
	public Camera getCamera() {
		return camera;
	}

	/**
	 * Default implementation simply calls render on each Renderable.
	 * 
	 * @see com.emarcotte.glfoo.Renderer#render(GraphicsDriver)
	 * @param driver
	 *            The GraphicsDriver that is currently running.
	 */
	@Override
	public void render(GraphicsDriver driver) {
		for (E r : renderables) {
			render(driver, r);
		}
	}

	/**
	 * Default implementation does nothing.
	 * 
	 * @see Renderer#initialize(GraphicsDriver)
	 * @param driver
	 *            The GraphicsDriver that is currently running.
	 */
	@Override
	public void initialize(GraphicsDriver driver) {

	}

	/**
	 * When a new renderable is added it need to be initialized and placed into
	 * the regular set of renderables for drawing.
	 * 
	 * @param driver
	 *            The GraphicsDriver that is currently running.
	 */
	protected void initializeNewRenderables(GraphicsDriver driver) {
		for (E r : newRenderables) {
			r.init(driver);
			renderables.add(r);
		}
		newRenderables.clear();
	}

	/**
	 * Adds a Renderable.
	 * 
	 * @param r
	 *            The renderable to add.
	 */
	public void addRenderable(E r) {
		newRenderables.add(r);
	}

	/**
	 * Removes a Renderable.
	 * 
	 * @param r
	 *            The renderable to remove.
	 */
	public void removeRenderable(E r) {
		renderables.remove(r);
	}

	/**
	 * Applies the camera perspective to the projection matrix and position to
	 * the modelview matrix..
	 * 
	 * This implementation also calls into initializeNewRenderables() which will
	 * be important should this method be overrode.
	 * 
	 * @see com.emarcotte.glfoo.Renderer#render(com.emarcotte.glfoo.GraphicsDriver)
	 * @param driver
	 *            The GraphicsDriver that is currently running.
	 */
	@Override
	public void renderBefore(GraphicsDriver driver) {

		initializeNewRenderables(driver);

		GL gl = driver.getGl();

		gl.glMatrixMode(GL.GL_PROJECTION);
		gl.glLoadIdentity();

		camera.applyPerspective(driver);

		gl.glMatrixMode(GL.GL_MODELVIEW);
		gl.glLoadIdentity();

		gl.glClearColor(1, 1, 1, 1);
		gl.glColor3d(.4, .4, .4);

		camera.applyPosition(driver);
	}

	/**
	 * Default implementation. Simply does glFlush().
	 * 
	 * @see com.emarcotte.glfoo.Renderer#renderAfter(com.emarcotte.glfoo.GraphicsDriver)
	 * @param driver
	 *            The GraphicsDriver that is currently running.
	 */
	@Override
	public void renderAfter(GraphicsDriver driver) {
		GL gl = driver.getGl();
		gl.glFlush();
	}

	/**
	 * Renders a Renderable. Override this with any specific setup that needs to
	 * be done for each Renderable.
	 * 
	 * @param driver
	 *            The GraphicsDriver that is currently running.
	 * @param r
	 *            The renderable to draw.
	 * 
	 */
	public abstract void render(GraphicsDriver driver, E r);

}
