package com.emarcotte.glfoo.implementation;

import java.awt.BorderLayout;
import java.awt.Dialog;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.lang.Thread.UncaughtExceptionHandler;

import javax.swing.JButton;
import javax.swing.JDialog;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextPane;
import javax.swing.SwingUtilities;

final class ExceptionDialog implements UncaughtExceptionHandler {
	private JDialog dialog;
	private Throwable t;
	private boolean finished = false;
	private boolean running = false;
	private Runnable launchAction;
	private Runnable disposeAction;
	private String title;

	private final class DialogCloser implements ActionListener {
		@Override
		public void actionPerformed(ActionEvent e) {
			ExceptionDialog.this.dispose();
		}
	}

	private final class DialogRunner implements Runnable {
		public void run() {
			ExceptionDialog.this.generateDialog();
		}
	}

	private static final String generateMessage(final Throwable t) {
		String message = String.format("<table width=\"100%%\">"
				+ "<tr><td>Caught</td>" + "<td>%1s</td></tr>\n", t.getClass()
				.toString());

		boolean dark = true;
		for (StackTraceElement e : t.getStackTrace()) {
			String color = dark ? "#eeeeee" : "#dddddd";
			dark = !dark;
			message += String.format("<tr style=\"background-color: %1s;\">\n"
					+ "<td></td><td> %2s in %3s at line %4d</td></tr>\n",
					color, e.getClassName(), e.getMethodName(), e
							.getLineNumber());
		}

		message += "</table>";

		if (t.getCause() != null) {
			message += generateMessage(t.getCause());
		}

		return message;
	}

	private final void createDialog() {
		JLabel label = new JLabel();
		JTextPane area = new JTextPane();
		JPanel panel = new JPanel();
		JScrollPane pane = new JScrollPane(area);
		String message = generateMessage(t);

		panel.add(label);
		label.setHorizontalTextPosition(JLabel.CENTER);
		label.setText(String.format("Exception: %1s", t.getMessage()));

		area.setContentType("text/html");
		area.setText("<html><body>" + message + "</body></html>");
		area.setEditable(false);

		dialog.add(panel, BorderLayout.NORTH);
		dialog.add(pane, BorderLayout.CENTER);

	}

	private final void createNullThrowableDialog() {
		JPanel panel = new JPanel();
		panel.add(new JLabel("Unknown exception caught"));
		dialog.add(panel);
	}

	protected void generateDialog() {
		dialog = new JDialog();
		dialog.setTitle(title);

		if (t == null) {
			createNullThrowableDialog();
		} else {
			createDialog();
		}

		JButton button = new JButton("Ok");
		JPanel buttonPanel = new JPanel();
		buttonPanel.add(button);
		button.addActionListener(new DialogCloser());
		dialog.add(buttonPanel, BorderLayout.SOUTH);
		dialog.setModalityType(Dialog.ModalityType.APPLICATION_MODAL);
		dialog.setDefaultCloseOperation(JDialog.DISPOSE_ON_CLOSE);
		dialog.pack();
		dialog.setVisible(true);
	}

	protected void dispose() {
		if (disposeAction != null) {
			disposeAction.run();
		}

		dialog.dispose();
		finished = true;
		dialog = null;
	}

	protected void launch(Thread thread, Throwable t) {
		this.t = t;
		this.running = true;
		if (!SwingUtilities.isEventDispatchThread()) {
			try {
				SwingUtilities.invokeAndWait(new DialogRunner());
			} catch (Exception e) {
				e.printStackTrace();
			}
		} else {
			generateDialog();
		}
	}

	public boolean isFinished() {
		return finished;
	}

	public boolean isRunning() {
		return running;
	}

	public void setLaunchAction(Runnable action) {
		this.launchAction = action;
	}

	public void setDisposeAction(Runnable action) {
		this.disposeAction = action;
	}

	public void setTitle(String title) {
		this.title = title;
	}

	@Override
	public void uncaughtException(Thread thread, Throwable t) {
		if (launchAction != null) {
			launchAction.run();
		}
		launch(thread, t);
	}

}
