package com.emarcotte.glfoo.implementation;

import com.emarcotte.glfoo.Camera;
import com.emarcotte.glfoo.GraphicsDriver;
import com.emarcotte.glfoo.math.Vector3;

/**
 * FocusableCamera is a {@link Camera} that is able to "look" at a given point.
 * It is also able to rotate around that point, however, other than the
 * rotation, the position is relatively fixed.
 * 
 * The mechanism employed is essentially spherical coordinates based on the
 * equations found at http://mathworld.wolfram.com/SphericalCoordinates.html
 * 
 * One key difference is of course the axis where rotations occurs and also the
 * axis which is up. In the case of this camera, y is "up", and z is depth.
 * 
 * @author Eugene Marcotte <emarcotte@gmail.com>
 */
public class FocusableCamera implements Camera {
	private final static double NINETY = Math.toRadians(90);

	/**
	 * The focus point.
	 */
	private Vector3 focusPoint;
	private Vector3 position;
	private Vector3 up;

	/**
	 * Near cutoff value.
	 */
	private double near;

	/**
	 * Far cutoff value.
	 */
	private double far;
	/**
	 * Perspective angle.
	 */
	private double perspective;

	/**
	 * Distance from focus point.
	 */
	private double distance;

	/**
	 * Amount of rotation about the Y (Primarily affects the X coordinate) axis,
	 * stored in radians.
	 */
	private double xAngle;

	/**
	 * Amount of rotation about the X (Primarily affects the Y coordinate) axis,
	 * stored in radians.
	 */
	private double yAngle;
	private double cosX;
	private double sinX;
	private double cosY;
	private double sinY;

	/**
	 * Set up the camera with a focus point at 0, 0, 0, at a distance of 5 with
	 * a far cut off of 100, near cut off of .01, a perspective of 45 and a
	 * focus orientation straight down the z axis.
	 */
	public FocusableCamera() {
		focusPoint = new Vector3();
		position = new Vector3();
		up = new Vector3();
		setDistance(5);
		setFocus(0, 0, 0);
		setFar(1000);
		setNear(.01);
		setPerspective(45);
		setXAngle(0);
		setYAngle(0);
	}

	/**
	 * Set the new X angle rotation (rotation around the Y axis).
	 * 
	 * @param angle
	 */
	public void setXAngle(double angle) {
		xAngle = Math.toRadians(angle);
		updateXAngle();
	}

	public void setYAngle(double angle) {
		yAngle = NINETY - Math.toRadians(angle);
		updateYAngle();
	}

	protected void updateXAngle() {
		cosX = Math.cos(xAngle);
		sinX = Math.sin(xAngle);
	}

	protected void updateYAngle() {
		cosY = Math.cos(yAngle);
		sinY = Math.sin(yAngle);
	}

	public void setFocus(double x, double y, double z) {
		setFocusX(x);
		setFocusY(y);
		setFocusZ(z);
	}

	public double getFocusX() {
		return focusPoint.getI();
	}

	public void setFocusX(double focusX) {
		focusPoint.setI(focusX);
	}

	public double getFocusY() {
		return focusPoint.getJ();
	}

	public void setFocusY(double focusY) {
		focusPoint.setJ(focusY);
	}

	public double getFocusZ() {
		return focusPoint.getK();
	}

	public void setFocusZ(double focusZ) {
		focusPoint.setK(focusZ);
	}

	public void addYAngle(double y) {
		yAngle += Math.toRadians(y);
		updateYAngle();
	}

	public void addXAngle(double x) {
		xAngle += Math.toRadians(x);
		updateXAngle();
	}

	public void setFocus(Vector3 position) {
		focusPoint = position;
	}

	public double getNear() {
		return near;
	}

	public void setNear(double near) {
		this.near = near;
	}

	public double getPerspective() {
		return perspective;
	}

	public void setPerspective(double perspective) {
		this.perspective = perspective;
	}

	public double getFar() {
		return far;
	}

	public void setFar(double far) {
		this.far = far;
	}

	public double getDistance() {
		return distance;
	}

	public void setDistance(double distance) {
		this.distance = distance;
	}

	public Vector3 getFocus() {
		return focusPoint;
	}

	public double getXAngle() {
		return xAngle;
	}

	public double getYAngle() {
		return NINETY - yAngle;
	}

	public Vector3 getPosition() {
		return new Vector3();
	}

	public Vector3 getUpVector() {
		return new Vector3();
	}

	@Override
	public void applyPerspective(GraphicsDriver driver) {
		driver.getGlu().gluPerspective(perspective, driver.getRatio(), near,
				far);
	}

	public void recalculateVectors() {
		position.setI(focusPoint.getI() + distance * sinX * sinY);
		position.setJ(focusPoint.getJ() + distance * cosY);
		position.setK(focusPoint.getK() + distance * cosX * sinY);

		up.setI(-sinX * cosY);
		up.setJ(sinY);
		up.setK(-cosX * cosY);

	}

	@Override
	public void applyPosition(GraphicsDriver driver) {
		// have to recalculate each frame since the focus vector may be owned by
		// something
		// outside of the camera.
		recalculateVectors();
		driver.getGlu().gluLookAt(position.getI(), position.getJ(),
				position.getK(), getFocusX(), getFocusY(), getFocusZ(),
				up.getI(), up.getJ(), up.getK());

	}
}
