package com.emarcotte.glfoo.implementation;

import java.awt.Point;
import java.awt.event.MouseEvent;
import java.nio.IntBuffer;
import java.util.HashMap;

import javax.media.opengl.GL;
import javax.media.opengl.glu.GLU;

import com.emarcotte.glfoo.Camera;
import com.emarcotte.glfoo.GraphicsDriver;
import com.emarcotte.glfoo.MouseHandler;
import com.emarcotte.glfoo.Renderer;
import com.sun.opengl.util.BufferUtil;

/**
 * Selection renderer draws things that can be 'selected' by the mouse using the
 * name buffer and then dispatches events to {@link Selectable} objects.
 * 
 * @author Eugene Marcotte <emarcotte@gmail.com>
 */
public class SelectionRenderer implements Renderer, MouseHandler {

	/**
	 * The size of the selection buffer. This could cause issues if it is too
	 * small, eventually this may drive it to be configurable.
	 */
	private final static int BUFFER_SIZE = 64;

	/**
	 * The name -> Selectable table. Each Selectable is assigned a name and when
	 * hits are detected on a name this table can be used to find which object
	 * it belongs to.
	 */
	private HashMap<Integer, Selectable> selectables;

	/**
	 * The selectable that is selected in the current frame.
	 */
	private Selectable currentFrameSelected;

	/**
	 * The selection buffer.
	 */
	private IntBuffer selectBuffer;

	/**
	 * The viewport information.
	 */
	private IntBuffer viewPort;

	/**
	 * The mouse position.
	 */
	private Point mousePoint;

	/**
	 * The camera to use when rendering in selection mode.
	 */
	private Camera camera;

	/**
	 * The last name that was generated.
	 * 
	 * TODO: Need to perhaps use a list of next names, which always has at least
	 * one item in it, or any unregistered Selectables.
	 */
	private int lastName;

	/**
	 * The number of hits in the hit buffer.
	 */
	private int hits;

	/**
	 * Creates a SelectionRenderer using an {@link OrthographicCamera}.
	 * Important note: on some broken hardware configurations the orthographic
	 * camera may be the only one that works for the SelectionRenderer.
	 */
	public SelectionRenderer() {
		this(new OrthographicCamera());
	}

	/**
	 * Creates a SelectionRenderer with an arbitrary Camera. The
	 * 
	 * @param c
	 */
	public SelectionRenderer(Camera c) {
		viewPort = IntBuffer.allocate(4);

		selectBuffer = BufferUtil.newIntBuffer(BUFFER_SIZE);
		mousePoint = new Point();
		selectables = new HashMap<Integer, Selectable>();
		lastName = 1;
		camera = c;
	}

	@Override
	public Camera getCamera() {
		return camera;
	}

	@Override
	public void initialize(GraphicsDriver driver) {
		GL gl = driver.getGl();
		gl.glSelectBuffer(BUFFER_SIZE, selectBuffer);
	}

	@Override
	public void renderBefore(GraphicsDriver driver) {
		GL gl = driver.getGl();
		GLU glu = driver.getGlu();

		gl.glGetIntegerv(GL.GL_VIEWPORT, viewPort);

		gl.glRenderMode(GL.GL_SELECT);
		gl.glInitNames();

		gl.glMatrixMode(GL.GL_PROJECTION);
		gl.glPushMatrix();
		gl.glLoadIdentity();

		glu.gluPickMatrix(mousePoint.x,
				(double) viewPort.get(3) - mousePoint.y, 1.0, 1.0, viewPort);

		camera.applyPerspective(driver);

		gl.glMatrixMode(GL.GL_MODELVIEW);
		gl.glLoadIdentity();
		camera.applyPosition(driver);
	}

	@Override
	public void render(GraphicsDriver driver) {
		for (Selectable r : selectables.values()) {
			render(driver, r);
		}
	}

	public void render(GraphicsDriver driver, Selectable r) {
		GL gl = driver.getGl();
		gl.glPushName(r.getName());
		r.getRenderable().draw(driver);
		gl.glPopName();
	}

	@Override
	public void renderAfter(GraphicsDriver driver) {
		GL gl = driver.getGl();
		gl.glMatrixMode(GL.GL_PROJECTION);
		gl.glPopMatrix();
		gl.glMatrixMode(GL.GL_MODELVIEW);
		hits = gl.glRenderMode(GL.GL_RENDER);

		dispatch(hits);
	}

	public void registerSelectable(Selectable item) {
		item.setName(lastName);
		selectables.put(lastName, item);
		lastName++;
	}

	public void unregisterSelectable(Selectable item) {
		selectables.remove(item);
	}

	public void dispatch(int hits) {
		if (hits == 0) {
			if (currentFrameSelected != null) {
				onLeave(currentFrameSelected);
			}
			currentFrameSelected = null;
			return;
		}

		int index = 0;
		int numberOfNames;
		int currentDepth;
		int closestDepth = selectBuffer.get(1);
		int currentName = selectBuffer.get(3);

		for (int i = 0; i < hits; i++) {
			numberOfNames = selectBuffer.get(index);
			index++;
			currentDepth = selectBuffer.get(index);
			index += 2;

			// Only get the first index (I can't think of a reason to get
			// others).
			if (currentDepth < closestDepth) {
				currentName = selectBuffer.get(index);
				closestDepth = currentDepth;
			}
			index += numberOfNames;
		}

		Selectable current = selectables.get(currentName);
		if (currentFrameSelected != null && currentFrameSelected != current) {
			onLeave(currentFrameSelected);
			onEnter(current);
		} else if (currentFrameSelected == null) {
			onEnter(current);
		}
		currentFrameSelected = current;
	}

	@Override
	public Point getMousePoint() {
		return mousePoint;
	}

	@Override
	public synchronized void mouseDragged(MouseEvent e) {
		if (currentFrameSelected != null) {
			double dx = e.getX() - mousePoint.x;
			double dy = e.getY() - mousePoint.y;
			onDrag(currentFrameSelected, dx, -dy);
		}

		mousePoint.x = e.getX();
		mousePoint.y = e.getY();
	}

	@Override
	public void mouseMoved(MouseEvent e) {
		mousePoint.x = e.getX();
		mousePoint.y = e.getY();
	}

	@Override
	public void mouseClicked(MouseEvent arg0) {
		if (currentFrameSelected != null) {
			onClick(currentFrameSelected);
		}
	}

	@Override
	public void mouseEntered(MouseEvent arg0) {

	}

	@Override
	public void mouseExited(MouseEvent arg0) {

	}

	@Override
	public void mousePressed(MouseEvent arg0) {
		if (currentFrameSelected != null) {
			onPressed(currentFrameSelected);
		}
	}

	@Override
	public void mouseReleased(MouseEvent arg0) {
		if (currentFrameSelected != null) {
			onRelease(currentFrameSelected);
		}
	}

	private static void onClick(Selectable s) {
		for (SelectableEventHandler e : s.getEventHandlers()) {
			if (e != null) {
				e.onClick();
			}
		}
	}

	private static void onDrag(Selectable s, double x, double y) {
		for (SelectableEventHandler e : s.getEventHandlers()) {
			e.onDrag(x, y);
		}
	}

	private static void onEnter(Selectable s) {
		for (SelectableEventHandler e : s.getEventHandlers()) {
			e.onEnter();
		}
	}

	private static void onLeave(Selectable s) {
		for (SelectableEventHandler e : s.getEventHandlers()) {
			e.onLeave();
		}
	}

	private static void onPressed(Selectable s) {
		for (SelectableEventHandler e : s.getEventHandlers()) {
			e.onPressed();
		}
	}

	private static void onRelease(Selectable s) {
		for (SelectableEventHandler e : s.getEventHandlers()) {
			e.onReleased();
		}
	}
}
