package com.emarcotte.glfoo.math;

/**
 * Representation of a Quaternion.
 * 
 * For more info on this stuff look to http://en.wikipedia.org/wiki/Quaternions
 * and http://mathworld.wolfram.com/Quaternion.html
 * 
 * In general I tried to design this in such a way that it would not have to new
 * objects often. This means that for some operation that would return a vector
 * or quat etc there will be two methods, one that news (and returns the result
 * via a return value) an object and one that takes a result object argument and
 * returns via that object.
 * 
 * @author Eugene Marcotte <emarcotte@gmail.com>
 */
public class Quaternion {

	/**
	 * The components of this Quaternion are stored in an array, always size 4.
	 */
	private double q[];

	/**
	 * X Component index.
	 */
	final static int X = 0;

	/**
	 * Y Component index.
	 */
	final static int Y = 1;

	/**
	 * Z Component index.
	 */
	final static int Z = 2;

	/**
	 * W Component index.
	 */
	final static int W = 3;

	/**
	 * Create a Quaternion as {[0, 0, 0], 1}.
	 */
	public Quaternion() {
		this(0, 0, 0, 1);
	}

	/**
	 * Create a Quaternion as {[x, y, z], 0}.
	 * 
	 * @param x
	 *            x component.
	 * @param y
	 *            y component.
	 * @param z
	 *            z component.
	 */
	public Quaternion(double x, double y, double z) {
		this(x, y, z, 0);
	}

	/**
	 * 
	 * Create a Quaternion as {[x, y, z], w}.
	 * 
	 * @param x
	 *            x component.
	 * @param y
	 *            y component.
	 * @param z
	 *            z component.
	 * @param w
	 *            w component.
	 */
	public Quaternion(double x, double y, double z, double w) {
		q = new double[4];
		setX(x);
		setY(y);
		setZ(z);
		setW(w);
	}

	/**
	 * Create a Quaternion from a euler axis angle (a vector and an amount of
	 * rotation around that vector).
	 * 
	 * @param axis
	 *            The axis to revolve around.
	 * @param angle
	 *            The amount to revolve, in degrees.
	 */
	public Quaternion(Vector3 axis, double angle) {
		this();
		setFromEulerAxisAngle(axis, angle);
	}

	/**
	 * Creates a copy of a Quaternion.
	 * 
	 * @param quaternion
	 *            Quaternion to copy.
	 */
	public Quaternion(Quaternion quaternion) {
		this(quaternion.q[X], quaternion.q[Y], quaternion.q[Z], quaternion.q[W]);
	}

	public void setFromEulerAxisAngle(Vector3 axis, double angle) {
		Vector3 n = new Vector3(axis);
		n.normalize();
		setFromEulerAxisAngle(n.getI(), n.getJ(), n.getK(), angle);
	}

	public void setFromEulerAxisAngle(double x, double y, double z, double angle) {
		double angleDividedByTwo = Math.toRadians(angle) / 2;
		double angleSin = Math.sin(angleDividedByTwo);
		setX(x * angleSin);
		setY(y * angleSin);
		setZ(z * angleSin);
		setW(Math.cos(angleDividedByTwo));
	}

	public void setX(double d) {
		q[X] = d;
	}

	public void setY(double d) {
		q[Y] = d;
	}

	public void setZ(double d) {
		q[Z] = d;
	}

	public void setW(double d) {
		q[W] = d;
	}

	public double getX() {
		return q[X];
	}

	public double getY() {
		return q[Y];
	}

	public double getZ() {
		return q[Z];
	}

	public double getW() {
		return q[W];
	}

	public double getMagnitude() {
		return Math.sqrt((q[X] * q[X]) + (q[Y] * q[Y]) + (q[Z] * q[Z])
				+ (q[W] * q[W]));
	}

	public void normalize() {
		division(getMagnitude());

	}

	public void division(double scalar) {
		if (scalar != 0.0) {
			q[W] /= scalar;
			q[X] /= scalar;
			q[Y] /= scalar;
			q[Z] /= scalar;
		}
	}

	public Quaternion multiplication(Quaternion qb) {
		Quaternion out = new Quaternion();

		multiplication(qb, out);

		return out;
	}

	public void multiplication(Quaternion qb, Quaternion out) {

		out.q[W] = (q[W] * qb.q[W]) - (q[X] * qb.q[X]) - (q[Y] * qb.q[Y])
				- (q[Z] * qb.q[Z]);
		out.q[X] = (q[X] * qb.q[W]) + (q[W] * qb.q[X]) + (q[Y] * qb.q[Z])
				- (q[Z] * qb.q[Y]);
		out.q[Y] = (q[Y] * qb.q[W]) + (q[W] * qb.q[Y]) + (q[Z] * qb.q[X])
				- (q[X] * qb.q[Z]);
		out.q[Z] = (q[Z] * qb.q[W]) + (q[W] * qb.q[Z]) + (q[X] * qb.q[Y])
				- (q[Y] * qb.q[X]);

	}

	public Quaternion multiplication(Vector3 v) {
		Quaternion out = new Quaternion();

		multiplication(v, out);

		return out;
	}

	public void multiplication(Vector3 v, Quaternion out) {
		out.q[W] = -(q[X] * v.getI()) - (q[Y] * v.getJ()) - (q[Z] * v.getK());
		out.q[X] = (q[W] * v.getI()) + (q[Y] * v.getK()) - (q[Z] * v.getJ());
		out.q[Y] = (q[W] * v.getJ()) + (q[Z] * v.getI()) - (q[X] * v.getK());
		out.q[Z] = (q[W] * v.getK()) + (q[X] * v.getJ()) - (q[Y] * v.getI());
	}

	public Quaternion conjugate() {
		Quaternion conjugate = new Quaternion();

		conjugate(conjugate);

		return conjugate;
	}

	public void conjugate(Quaternion conjugate) {
		conjugate.q[X] = -q[X];
		conjugate.q[Y] = -q[Y];
		conjugate.q[Z] = -q[Z];
		conjugate.q[W] = q[W];
	}

	/**
	 * Takes some given vector and rotates it by this quaternion. This is
	 * accomplished by getting the conjugate, normalizing it then multiplying
	 * this quaternion by the vector, and that result by the conjugate: q.x.q* =
	 * result.
	 * 
	 * @param v
	 *            The point to rotate
	 * @return The results.
	 */
	public Vector3 rotatePoint(Vector3 v) {
		Vector3 out = new Vector3();
		rotatePoint(v, out);
		return out;
	}

	/**
	 * Takes some given vector and rotates it by this quaternion. This is
	 * accomplished by getting the conjugate, normalizing it then multiplying
	 * this quaternion by the vector, and that result by the conjugate: q.x.q* =
	 * result.
	 * 
	 * @param v
	 *            The point to rotate
	 * @param out
	 *            Where to store the results.
	 */
	public void rotatePoint(Vector3 v, Vector3 out) {
		Quaternion con = conjugate();

		con.normalize();

		Quaternion rotate_result = multiplication(v).multiplication(con);
		out.setI(rotate_result.q[X]);
		out.setJ(rotate_result.q[Y]);
		out.setK(rotate_result.q[Z]);
	}

	public double dotProduct(Quaternion q1) {
		return (q[X] * q1.q[X]) + (q[Y] * q1.q[Y]) + (q[Z] * q1.q[Z]);
	}

	/**
	 * TODO: This is a mess.
	 * 
	 * @param original
	 * @param next
	 * @param factor
	 */
	public void slerp(Quaternion original, Quaternion next, double factor) {
		double cosom = (original.q[X] * next.q[X])
				+ (original.q[Y] * next.q[Y]) + (original.q[Z] * next.q[Z])
				+ (original.q[W] * next.q[W]);

		double sclp, sclq, omega, sinom;

		if ((cosom < 0.0)) {
			cosom = -cosom;
			q[X] = -next.q[X];
			q[Y] = -next.q[Y];
			q[Z] = -next.q[Z];
			q[W] = -next.q[W];
		} else {
			q[X] = next.q[X];
			q[Y] = next.q[Y];
			q[Z] = next.q[Z];
			q[W] = next.q[W];
		}

		omega = Math.acos(cosom);
		sinom = Math.sin(omega);
		sclp = Math.sin((1.0 - factor) * omega) / sinom;
		sclq = Math.sin(factor * omega) / sinom;
		q[X] = sclp * original.q[X] + sclq * q[X];
		q[Y] = sclp * original.q[Y] + sclq * q[Y];
		q[Z] = sclp * original.q[Z] + sclq * q[Z];
		q[W] = sclp * original.q[W] + sclq * q[W];

	}

	/**
	 * TODO: Verify?
	 * 
	 * @param original
	 * @param next
	 * @param factor
	 */
	public void normalizedLinearInterpolate(Quaternion original,
			Quaternion next, double factor) {
		q[X] = original.q[X] + factor * (next.q[X] - original.q[X]);
		q[Y] = original.q[Y] + factor * (next.q[Y] - original.q[Y]);
		q[Z] = original.q[Z] + factor * (next.q[Z] - original.q[Z]);
		q[W] = original.q[W] + factor * (next.q[W] - original.q[W]);
		normalize();
	}

	public void copy(Quaternion q) {
		setX(q.getX());
		setY(q.getY());
		setZ(q.getZ());
		setW(q.getW());
	}

}
