package com.emarcotte.glfoo.utils;

import java.io.IOException;
import java.io.InputStream;
import java.nio.IntBuffer;

import javax.media.opengl.GL;
import javax.media.opengl.glu.GLU;

import com.sun.opengl.util.BufferUtil;
import com.sun.opengl.util.StreamUtil;

/**
 * Wrapper class for GLSL vertex and fragment shaders. Shaders are compiled then
 * added to a shader program which can be used to perform custom rendering
 * through opengl.
 * 
 * @author Eugene Marcotte
 */
public class GLShader {

	/**
	 * OpenGL's id for this shader.
	 */
	private int shaderId;

	/**
	 * The type of shader, this should be one of GL_VERTEX_SHADER, or
	 * GL_VRAGMENT_SHADER -- any other stuff may work, but was not intended.
	 */
	private int shaderType;

	/**
	 * The GL that this shader is on.
	 */
	private GL gl;

	/**
	 * Hidden constructor to enforce factory construction since calls to OpenGL
	 * are required to ensure that the shader is built properly.
	 * 
	 * @param gl
	 *            The GL context that the shader is on.
	 * @param shaderId
	 *            The OpenGL id of the shader.
	 * @param shaderType
	 *            The type of the shader.
	 */
	private GLShader(GL gl, int shaderId, int shaderType) {
		this.gl = gl;
		this.shaderId = shaderId;
		this.shaderType = shaderType;
	}

	/**
	 * Factory method for creating a shader. Enforces valid shader's are used
	 * when generating new instances.
	 * 
	 * TODO: Make it so that source can be passed in rather than resource paths.
	 * 
	 * @param gl
	 *            An OpenGL context to use.
	 * @param glu
	 *            A GLU context to use.
	 * @param shaderSourcePath
	 *            The path to the shader. Note that this is not the source of
	 *            the shader but the path to it. TODO: Fix for just source.
	 * @param shaderType
	 *            The type of shader to create (vertex, or fragment).
	 * @return A fully constructed shader object or null if there was a failure.
	 *         TODO: Move to exceptions.
	 */
	public static GLShader load(GL gl, GLU glu, String shaderSourcePath,
			int shaderType) {

		InputStream inputStream = ResourceHandler.getHandler()
				.getResourceAsStream(shaderSourcePath);

		if (inputStream == null) {
			return null;
		}

		try {
			byte[] file = StreamUtil.readAll(inputStream);

			String lines = new String(file);

			int shaderId = gl.glCreateShader(shaderType);

			GLShader shader = new GLShader(gl, shaderId, shaderType);
			shader.compileShader(glu, new String[] { lines });

			return shader;

		} catch (IOException e) {
			e.printStackTrace();
			return null;
		}
	}

	/**
	 * Utility method to compile the source of the shader. Note that splitting a
	 * shader's source lines up is not necessary, but passing them in as a
	 * string array is. That is, you can have an array of size 1 with a very
	 * long string filled with '\n's.
	 * 
	 * TODO: This current spits out information to stdout. It probably should
	 * not.
	 * 
	 * @param gl
	 *            a GL context to use.
	 * @param glu
	 *            A GLU context to use.
	 * @param lines
	 *            The lines of the source file.
	 * @return true or false depending upon if the source compiled or not.
	 */
	private boolean compileShader(GLU glu, String[] lines) {
		gl.glShaderSource(shaderId, lines.length, lines, null);
		gl.glCompileShader(shaderId);

		if (!GLUtil.checkError(gl, glu)) {
			return false;
		}

		IntBuffer status = BufferUtil.newIntBuffer(1);

		gl.glGetShaderiv(shaderId, GL.GL_COMPILE_STATUS, status);

		System.out.println(status.get() == GL.GL_TRUE ? Messages
				.getString("GLShader.Compile_Successful") //$NON-NLS-1$
				: Messages.getString("GLShader.Compile_Not_Successful")); //$NON-NLS-1$

		IntBuffer iVal = BufferUtil.newIntBuffer(1);

		gl.glGetShaderiv(shaderId, GL.GL_INFO_LOG_LENGTH, iVal);

		int length = iVal.get();

		if (length <= 1) {
			return true;
		}

		byte[] infoBytes = new byte[length];
		gl.glGetShaderInfoLog(shaderId, length, null, 0, infoBytes, 0);

		System.out.format(Messages.getString("GLShader.Compile_Log"), shaderId);
		System.out.println(new String(infoBytes));

		return false;
	}

	public int getShaderId() {
		return shaderId;
	}

	public int getShaderType() {
		return shaderType;
	}
}
