package com.emarcotte.glfoo.utils;

import java.util.HashMap;
import java.util.Map;

import javax.media.opengl.GL;

/**
 * This class only works for ONE GL context.
 * 
 * http://www.mevis.de/opengl/glLight.html
 * 
 * Float/int controls for: GL_SPOT_EXPONENT, GL_SPOT_CUTOFF,
 * GL_CONSTANT_ATTENUATION, GL_LINEAR_ATTENUATION, GL_QUADRATIC_ATTENUATION.
 * 
 * Vector Controls for: GL_AMBIENT, GL_DIFFUSE, GL_SPECULAR, GL_POSITION,
 * GL_SPOT_DIRECTION, GL_SPOT_EXPONENT, GL_SPOT_CUTOFF, GL_CONSTANT_ATTENUATION,
 * GL_LINEAR_ATTENUATION, GL_QUADRATIC_ATTENUATION.
 * 
 * 
 * Plans:
 * 
 * Keep a static hashmap:
 * 
 * [GL_LIGHTi] => {some Light object, boolean enabled, }
 * 
 * If an object wants to enable a light, it looks if the light is bound to a
 * target.
 * 
 * If it is currently bound, it simply enables the light. If the light is not
 * bound, Light will search for a light target that is not enabled and bind to
 * it.
 * 
 * 
 * @author Eugene Marcotte
 * 
 */
public class Light {

	/**
	 * The light position. Doubles as direction if it is a directional light.
	 * 
	 * position[3] = (!directional)
	 */
	private float[] position;

	private float[] spotlightDirection;

	private float[] ambient;

	private float[] diffuse;

	private float[] specular;

	private boolean update;

	private GL gl;

	private static HashMap<Integer, LightEntry> lightTargets;

	private static HashMap<Light, Integer> boundLights;

	static {
		lightTargets = new HashMap<Integer, LightEntry>();
		boundLights = new HashMap<Light, Integer>();

		int numLights = Math.min(GL.GL_MAX_LIGHTS, 8);
		for (int i = GL.GL_LIGHT0; i < GL.GL_LIGHT0 + numLights; i++) {
			lightTargets.put(i, new LightEntry());
		}
	}

	/**
	 * Creates a directional light position = { 0, 1, 0 } ambient = {0, 0, 0}
	 * diffuse = {0, 0, 0} specular = {1, 1, 1}
	 */
	public Light(GL gl) {
		this.gl = gl;
		position = new float[] { 0, 1, 0, 0 };
		ambient = new float[] { 0, 0, 0, 1 };
		diffuse = new float[] { 0, 0, 0, 1 };
		specular = new float[] { 1, 1, 1, 1 };
		spotlightDirection = new float[] { 0, -1, 0 };
	}

	public void setAmbientColor(float r, float g, float b) {
		update = true;
		ambient[0] = r;
		ambient[1] = g;
		ambient[2] = b;
	}

	public void setDiffuseColor(float r, float g, float b) {
		update = true;
		diffuse[0] = r;
		diffuse[1] = g;
		diffuse[2] = b;
	}

	public void setSpecularColor(float r, float g, float b) {
		update = true;
		specular[0] = r;
		specular[1] = g;
		specular[2] = b;
	}

	public void setPosition(float x, float y, float z) {
		update = true;
		position[0] = x;
		position[1] = y;
		position[2] = z;
	}

	private void applyLight(int lightTarget) {
		gl.glLightfv(lightTarget, GL.GL_AMBIENT, ambient, 0);
		gl.glLightfv(lightTarget, GL.GL_DIFFUSE, diffuse, 0);
		gl.glLightfv(lightTarget, GL.GL_SPECULAR, specular, 0);
		gl.glLightfv(lightTarget, GL.GL_POSITION, position, 0);
		gl.glLightfv(lightTarget, GL.GL_SPOT_DIRECTION, spotlightDirection, 0);
	}

	/**
	 * Toggle for light type. Directional lights are NOT spotlight type lights
	 * and shine in one direction (from infinity)
	 * 
	 * If a light is not directional, it is a point light. Point lights are set
	 * at some point. All point lights are spotlights, but by default have a
	 * cone of 180degrees.
	 */
	public void setDirectional(boolean directional) {
		position[3] = directional ? 0 : 1;
	}

	public void disable(GL gl) {
		Integer target = boundLights.get(this);
		if (target != null) {
			lightTargets.get(target).setEnabled(false);
			gl.glDisable(target);
		}
	}

	public void enable() {
		// If this light is not bound, look for an empty light
		Integer target = boundLights.get(this);
		if (target == null) {
			for (Map.Entry<Integer, LightEntry> entry : lightTargets.entrySet()) {
				if (!entry.getValue().isEnabled()) {
					boundLights.remove(entry.getValue().getLight());
					entry.getValue().setLight(this);
					entry.getValue().setEnabled(true);
					applyLight(entry.getKey());
					update = false;
				}
			}
		} else if (this.update) {
			applyLight(target);
			update = false;
		}
	}
}
