package com.emarcotte.glfoo.utils;

import javax.media.opengl.GL;

/**
 * This class only works for ONE GL context.
 * 
 * http://www.mevis.de/opengl/glLight.html
 * 
 * Float/int controls for: GL_SPOT_EXPONENT, GL_SPOT_CUTOFF,
 * GL_CONSTANT_ATTENUATION, GL_LINEAR_ATTENUATION, GL_QUADRATIC_ATTENUATION.
 * 
 * Vector Controls for: GL_AMBIENT, GL_DIFFUSE, GL_SPECULAR, GL_POSITION,
 * GL_SPOT_DIRECTION, GL_SPOT_EXPONENT, GL_SPOT_CUTOFF, GL_CONSTANT_ATTENUATION,
 * GL_LINEAR_ATTENUATION, GL_QUADRATIC_ATTENUATION.
 * 
 * 
 * Plans:
 * 
 * Keep a static hashmap:
 * 
 * [GL_LIGHTi] => {some Light object, boolean enabled, }
 * 
 * If an object wants to enable a light, it looks if the light is bound to a
 * target.
 * 
 * If it is currently bound, it simply enables the light. If the light is not
 * bound, Light will search for a light target that is not enabled and bind to
 * it.
 * 
 * 
 * @author Eugene Marcotte
 * 
 */
public class Light2 {

	/**
	 * The light position. Doubles as direction if it is a directional light.
	 * 
	 * position[3] = (!directional)
	 */
	private float[] position;

	private float[] spotlightDirection;

	private float[] ambient;

	private float[] diffuse;

	private float[] specular;

	/**
	 * Creates a directional light position = { 0, 1, 0 } ambient = {0, 0, 0}
	 * diffuse = {0, 0, 0} specular = {1, 1, 1}
	 */
	public Light2() {
		position = new float[] { 0, 1, 0, 0 };
		ambient = new float[] { 0, 0, 0, 1 };
		diffuse = new float[] { 0, 0, 0, 1 };
		specular = new float[] { 1, 1, 1, 1 };
		spotlightDirection = new float[] { 0, -1, 0 };
	}

	public void setAmbientColor(float r, float g, float b) {
		ambient[0] = r;
		ambient[1] = g;
		ambient[2] = b;
	}

	public void setDiffuseColor(float r, float g, float b) {
		diffuse[0] = r;
		diffuse[1] = g;
		diffuse[2] = b;
	}

	public void setSpecularColor(float r, float g, float b) {
		specular[0] = r;
		specular[1] = g;
		specular[2] = b;
	}

	public void setPosition(float x, float y, float z) {
		position[0] = x;
		position[1] = y;
		position[2] = z;
	}

	/**
	 * Toggle for light type. Directional lights are NOT spotlight type lights
	 * and shine in one direction (from infinity)
	 * 
	 * If a light is not directional, it is a point light. Point lights are set
	 * at some point. All point lights are spotlights, but by default have a
	 * cone of 180degrees.
	 */
	public void setDirectional(boolean directional) {
		position[3] = directional ? 0 : 1;
	}

	public void applyLight(int lightTarget, GL gl) {
		gl.glLightfv(lightTarget, GL.GL_AMBIENT, ambient, 0);
		gl.glLightfv(lightTarget, GL.GL_DIFFUSE, diffuse, 0);
		gl.glLightfv(lightTarget, GL.GL_SPECULAR, specular, 0);
		gl.glLightfv(lightTarget, GL.GL_POSITION, position, 0);
		gl.glLightfv(lightTarget, GL.GL_SPOT_DIRECTION, spotlightDirection, 0);
	}

}
