package com.emarcotte.glfoo.utils;

import java.nio.DoubleBuffer;

import javax.media.opengl.GL;

import com.emarcotte.glfoo.GraphicsDriver;
import com.emarcotte.glfoo.Renderable;
import com.sun.opengl.util.BufferUtil;

/**
 * Initial attempt at a wrapper for various VBO bits.
 * 
 * Eventually it should be possible to do most operations through this object.
 * 
 * @author Eugene Marcotte <emarcotte@gmail.com>
 */
public class VertexBufferObject implements Renderable {
	protected int positionBuffer;
	protected int texCoordBuffer;
	protected int normalBuffer;
	protected DoubleBuffer verts;
	protected DoubleBuffer texCoords;
	protected DoubleBuffer normals;
	protected int faceMode, vertCount;
	protected int textureComponentCount;

	/**
	 * Set the normal data for this VBO. Only accepts 3 component normal data in
	 * the form {X1, Y1, Z1, X2, Y2, Z2,...}
	 * 
	 * The number of normals must be equal to the number of vertices.
	 * 
	 * @param n
	 *            The normal data.
	 * @throws IllegalArgumentException
	 *             When the number of normals is not valid.
	 */
	public void setNormalData(double[] n) {
		if (n.length != vertCount * 3) {
			throw new IllegalArgumentException(
					"Number of normals must equal number of vertices");
		}
		normals = DoubleBuffer.wrap(n);
	}

	/**
	 * Set the texture coordinates for the VBO. The number of coordinates must
	 * be equal to the number of vertices. As a result you must first set your
	 * vertex data before specifying texture coordinates.
	 * 
	 * @param t
	 *            The array of component positions in the format {X1, Y1, X2,
	 *            Y2, .. } assuming 2 components per coordinate.
	 * @param numComponents
	 *            The number of components that make up a coordinate.
	 * 
	 * @throws IllegalArgumentException
	 *             When number of coordinates or components is invalid.
	 */
	public void setTextureCoordinates(double[] t, int numComponents) {
		if (numComponents < 1 || numComponents > 3) {
			throw new IllegalArgumentException(
					"Number of components must be between 1 and 3");
		}

		if (t.length != vertCount * numComponents) {
			throw new IllegalArgumentException(
					"Number of texture coordinates must be equal to number of vertices");
		}

		texCoords = DoubleBuffer.wrap(t);
	}

	public void setVertexPositions(double[] v, int faceMode) {
		int size = 0;
		if (faceMode == GL.GL_TRIANGLES) {
			this.faceMode = faceMode;
			size = 3;
		} else if (faceMode == GL.GL_QUADS) {
			this.faceMode = faceMode;
			size = 4;
		} else {
			throw new IllegalArgumentException("Unsupported face mode");
		}
		verts = DoubleBuffer.wrap(v);
		vertCount = v.length / size;
	}

	@Override
	public void draw(GraphicsDriver driver) {
		GL gl = driver.getGl();

		gl.glEnableClientState(GL.GL_VERTEX_ARRAY);
		gl.glBindBufferARB(GL.GL_ARRAY_BUFFER, positionBuffer);
		gl.glBufferDataARB(GL.GL_ARRAY_BUFFER, verts.capacity()
				* BufferUtil.SIZEOF_DOUBLE, verts, GL.GL_STREAM_DRAW);
		gl.glVertexPointer(3, GL.GL_DOUBLE, 0, 0);

		if (texCoords != null) {
			gl.glEnableClientState(GL.GL_TEXTURE_COORD_ARRAY);
			gl.glBindBufferARB(GL.GL_ARRAY_BUFFER, texCoordBuffer);
			gl.glBufferDataARB(GL.GL_ARRAY_BUFFER, texCoords.capacity()
					* BufferUtil.SIZEOF_DOUBLE, texCoords, GL.GL_STREAM_DRAW);
			gl.glTexCoordPointer(2, GL.GL_DOUBLE, 0, 0);
		}

		if (normals != null) {
			gl.glEnableClientState(GL.GL_NORMAL_ARRAY);
			gl.glBindBufferARB(GL.GL_ARRAY_BUFFER, normalBuffer);
			gl.glBufferDataARB(GL.GL_ARRAY_BUFFER, normals.capacity()
					* BufferUtil.SIZEOF_DOUBLE, normals, GL.GL_STREAM_DRAW);
			gl.glNormalPointer(GL.GL_DOUBLE, 0, 0);
		}

		gl.glDrawArrays(faceMode, 0, vertCount);

		gl.glDisableClientState(GL.GL_NORMAL_ARRAY);
		gl.glDisableClientState(GL.GL_TEXTURE_COORD_ARRAY);
		gl.glDisableClientState(GL.GL_VERTEX_ARRAY);
	}

	@Override
	public void init(GraphicsDriver driver) {
		GL gl = driver.getGl();

		positionBuffer = GLUtil.genBuffer(gl);

		gl.glBindBufferARB(GL.GL_ARRAY_BUFFER, positionBuffer);
		gl.glBufferDataARB(GL.GL_ARRAY_BUFFER, verts.capacity()
				* BufferUtil.SIZEOF_DOUBLE, verts, GL.GL_STREAM_DRAW);

		if (texCoords != null) {
			texCoordBuffer = GLUtil.genBuffer(gl);
			gl.glBindBufferARB(GL.GL_ARRAY_BUFFER, texCoordBuffer);
			gl.glBufferDataARB(GL.GL_ARRAY_BUFFER, texCoords.capacity()
					* BufferUtil.SIZEOF_DOUBLE, texCoords, GL.GL_STREAM_DRAW);
		}

		if (normals != null) {
			normalBuffer = GLUtil.genBuffer(gl);
			gl.glBindBufferARB(GL.GL_ARRAY_BUFFER, normalBuffer);
			gl.glBufferDataARB(GL.GL_ARRAY_BUFFER, normals.capacity()
					* BufferUtil.SIZEOF_DOUBLE, normals, GL.GL_STREAM_DRAW);
		}

	}

}
