package com.emarcotte.testrio;

import java.io.FileNotFoundException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Scanner;
import java.util.regex.MatchResult;
import java.util.regex.Pattern;

import com.emarcotte.glfoo.math.Quaternion;
import com.emarcotte.glfoo.math.Vector3;

/**
 * MD5 Model loader.
 * 
 * @author Eugene Marcotte
 */
public class MD5Loader {

	/**
	 * The model that this loader loaded.
	 */
	private MD5Model model;

	/**
	 * 
	 */
	private MD5Mesh currentMesh;

	/**
	 * 
	 */
	private MD5Frame bindpos;

	/**
	 * The regex string for reading a line of joint data.
	 */
	private static final String MESH_JOINT_REGEX = "\\s\"(.+)\"\\s(-??\\d+)\\s\\("
			+ "\\s(-??\\d+(?:\\.\\d+)??)" // Match pos x
			+ "\\s(-??\\d+(?:\\.\\d+)??)" // Match pos y
			+ "\\s(-??\\d+(?:\\.\\d+)??)" // Match pos z
			+ "\\s\\)\\s\\(" // Parenthesis
			+ "\\s(-??\\d+(?:\\.\\d+)??)" // Match qX
			+ "\\s(-??\\d+(?:\\.\\d+)??)" // Match qY
			+ "\\s(-??\\d+(?:\\.\\d+)??)" // Match qZ
			+ "\\s\\).*";// Parenthesis

	private static final String MESH_VERT_REGEX = "\\svert\\s(\\d+)\\s\\("
			+ "\\s(-??\\d+(?:\\.\\d+)??)" // uvx
			+ "\\s(-??\\d+(?:\\.\\d+)??)" // uvy
			+ "\\s\\)\\s(\\d+)\\s(\\d+).*";

	private static final Pattern MESH_JOINT_PATTERN = Pattern
			.compile(MESH_JOINT_REGEX);

	private static final Pattern MESH_VERT_PATTERN = Pattern
			.compile(MESH_VERT_REGEX);

	/**
	 * Load things.
	 * http://wiki.doom3reference.com/wiki/MD5MESH_%28file_format%29
	 * 
	 * @param mesh
	 * @param animations
	 * 
	 * @throws FileNotFoundException
	 *             if any of the files were not found
	 */
	public void loadModel(InputStream mesh, Map<String, InputStream> animations) {

		Scanner meshScanner = new Scanner(mesh);

		int numMeshes, numJoints;

		// Read basic stuff
		meshScanner.findWithinHorizon("MD5Version", 0);
		int version = meshScanner.nextInt();

		meshScanner.findWithinHorizon("numJoints", 0);
		numJoints = meshScanner.nextInt();

		meshScanner.findWithinHorizon("numMeshes", 0);
		numMeshes = meshScanner.nextInt();

		model = new MD5Model(numJoints, numMeshes);

		System.out.println("Loading a version " + version + " model");
		System.out.println("Loading joints..");

		loadJoints(meshScanner, numJoints);
		System.out.println("Loading meshes..");

		loadMeshes(meshScanner, numMeshes);
		loadAnimations(animations);

		model.computeBindPosition();
		for (MD5Mesh m : model.meshes) {
			m.computeBindPositionNormals();
		}
	}

	/**
	 * Load data from a single joint.
	 * 
	 * Format: "Name" Parent (X Y Z) (Qx Qy Qz)
	 * 
	 * Example: "Orgin" -1 ( 0 0 0 ) ( -0.5 -0.5 -0.5 )
	 * 
	 * @param meshScanner
	 *            Scanner to read from.
	 * 
	 * @return a Joint successfully loaded.
	 */
	private MD5Joint loadJoint(Scanner meshScanner) {
		MatchResult result;

		if (meshScanner.findWithinHorizon(MESH_JOINT_PATTERN, 0) == null) {
			System.err.println("Expected another joint.");
		}

		result = meshScanner.match();

		String name = result.group(1);
		int parentId = Integer.parseInt(result.group(2));
		double x = Double.parseDouble(result.group(3));
		double y = Double.parseDouble(result.group(4));
		double z = Double.parseDouble(result.group(5));
		double qx = Double.parseDouble(result.group(6));
		double qy = Double.parseDouble(result.group(7));
		double qz = Double.parseDouble(result.group(8));

		MD5Joint parentJoint = null;

		if (parentId != -1) {
			parentJoint = bindpos.getJoint(parentId);
		}
		Vector3 position = new Vector3(x, y, z);
		Quaternion orientation = new Quaternion(qx, qy, qz);
		computeW(orientation);
		return new MD5Joint(name, parentJoint, position, orientation);
	}

	/**
	 * Load all joints.
	 * 
	 * @param meshScanner
	 * @param numJoints
	 */
	private void loadJoints(Scanner meshScanner, int numJoints) {

		meshScanner.findWithinHorizon("joint", 0);
		bindpos = new MD5Frame(numJoints);

		for (int i = 0; i < numJoints; i++) {
			MD5Joint j = loadJoint(meshScanner);
			bindpos.addJoint(i, j);
		}
		model.setBaseFrame(bindpos);
	}

	/**
	 * Load the UVCoords from the mesh file. These are called vertices in the
	 * file, however, they are pretty just reference to weights for UV
	 * coordinates.
	 * 
	 * @param meshScanner
	 *            Scanner to read from.
	 */
	private void loadMD5Verts(Scanner meshScanner) {
		int numVerts, vertId, weightIndex, weightElem;
		double textureU, textureV;
		MatchResult result;

		meshScanner.findWithinHorizon("numverts", 0);
		numVerts = meshScanner.nextInt();

		currentMesh.setNumVertices(numVerts);

		for (int i = 0; i < numVerts; i++) {
			if (meshScanner.findWithinHorizon(MESH_VERT_PATTERN, 0) == null) {
				System.err.println("Expected more verts");
			}

			result = meshScanner.match();

			vertId = Integer.parseInt(result.group(1));
			textureU = Double.parseDouble(result.group(2));
			textureV = Double.parseDouble(result.group(3));
			weightIndex = Integer.parseInt(result.group(4));
			weightElem = Integer.parseInt(result.group(5));

			currentMesh.addVertex(vertId, textureU, textureV, weightIndex,
					weightElem);
		}
	}

	private void loadTris(Scanner meshScanner) {
		int triId = 0, numTris;
		meshScanner.findWithinHorizon("numtris", 0);

		numTris = meshScanner.nextInt();
		currentMesh.setNumTriangles(numTris);

		while (triId < numTris - 1) {
			int vert1, vert2, vert3;
			meshScanner.findWithinHorizon("tri", 0);
			triId = meshScanner.nextInt();
			vert1 = meshScanner.nextInt();
			vert2 = meshScanner.nextInt();
			vert3 = meshScanner.nextInt();

			currentMesh.addTriangle(triId, vert1, vert2, vert3);
		}
	}

	private void loadWeights(Scanner meshScanner) {
		int numWeights, currentWeight = 0;

		meshScanner.findWithinHorizon("numweights", 0);
		numWeights = meshScanner.nextInt();
		currentMesh.setNumWeights(numWeights);

		while (currentWeight < numWeights - 1) {
			int jointIndex;
			double weightValue;
			double xpos, ypos, zpos;

			meshScanner.findWithinHorizon("weight", 0);

			currentWeight = meshScanner.nextInt();
			jointIndex = meshScanner.nextInt();
			weightValue = meshScanner.nextDouble();

			meshScanner.findWithinHorizon(" \\( ", 0);
			xpos = meshScanner.nextDouble();
			ypos = meshScanner.nextDouble();
			zpos = meshScanner.nextDouble();

			currentMesh.addWeight(currentWeight, xpos, ypos, zpos, weightValue,
					jointIndex);
		}

	}

	/**
	 * Load all the meshes.
	 * 
	 * @param meshScanner
	 *            Scanner to read from.
	 * @param numMeshes
	 */
	private void loadMeshes(Scanner meshScanner, int numMeshes) {
		for (int i = 0; i < numMeshes; i++) {

			currentMesh = new MD5Mesh();

			System.out.println("Loading verts..");
			loadMD5Verts(meshScanner);
			System.out.println("Loading tris..");
			loadTris(meshScanner);
			System.out.println("Loading weights...");
			loadWeights(meshScanner);
			System.out.println("Done with mesh.");
			model.addMesh(i, currentMesh);
		}
	}

	public MD5Model getModel() {
		return model;
	}

	private void loadAnimations(Map<String, InputStream> animations) {
		System.out.println("Loading an animation...");
		for (String name : animations.keySet()) {
			MD5Animation anim = loadAnimation(name, animations.get(name));
			model.addAnimation(anim);
		}

	}

	private MD5Animation loadAnimation(String name, InputStream stream) {
		Scanner animScanner = new Scanner(stream);
		System.out.println("File: " + name);
		// read header info
		System.out.println("Read header info..");
		animScanner.findWithinHorizon("numFrames", 0);
		int numFrames = animScanner.nextInt();
		animScanner.findWithinHorizon("numJoints", 0);
		int numJoints = animScanner.nextInt();
		animScanner.findWithinHorizon("frameRate", 0);
		int frameRate = animScanner.nextInt();
		animScanner.findWithinHorizon("numAnimatedComponents", 0);
		int numAnimComponents = animScanner.nextInt();

		// Create some data, and by some, I mean alot.
		System.out.println("Setup..");

		MD5Animation currentAnimation = new MD5Animation(name, frameRate);
		List<Quaternion> orientations = new ArrayList<Quaternion>(numJoints);
		List<MD5Frame> framesList = new ArrayList<MD5Frame>(numFrames);
		List<Vector3> positions = new ArrayList<Vector3>(numJoints);
		List<String> names = new ArrayList<String>(numJoints);
		int[][] jointFlags = new int[numJoints][];

		System.out.println("Read joint flags...");

		// Read joint flags
		for (int i = 0; i < numJoints; i++) {
			jointFlags[i] = new int[3];

			animScanner.findWithinHorizon("\"(.+)\"", 0);
			// Read the name...
			MatchResult result = animScanner.match();
			names.add(result.group(1));

			jointFlags[i][0] = animScanner.nextInt();
			jointFlags[i][1] = animScanner.nextInt();
			jointFlags[i][2] = animScanner.nextInt();
		}

		System.out.println("Load bounding box..");

		// TODO Get the stuff about bounding boxes here... Is it really useful?

		System.out.println("Load base frame info");

		animScanner.findWithinHorizon("baseframe \\{", 0);

		// Read the base frame info (joint pos) (orientation)
		for (int i = 0; i < numJoints; i++) {
			animScanner.findWithinHorizon("\\(", 0);
			double posX = animScanner.nextDouble();
			double posY = animScanner.nextDouble();
			double posZ = animScanner.nextDouble();
			positions.add(new Vector3(posX, posY, posZ));

			animScanner.findWithinHorizon("\\(", 0);
			double oX = animScanner.nextDouble();
			double oY = animScanner.nextDouble();
			double oZ = animScanner.nextDouble();

			Quaternion q = new Quaternion(oX, oY, oZ);
			computeW(q);
			orientations.add(q);
		}

		System.out.println("Process frames...");
		// frame processing
		for (int i = 0; i < numFrames; i++) {
			animScanner.findWithinHorizon("frame \\d* \\{", 0);

			double frameData[] = new double[numAnimComponents];

			for (int data = 0; data < numAnimComponents; data++) {
				frameData[data] = animScanner.nextDouble();
			}

			MD5Frame skel = buildSkeletonFrame(names, frameData, jointFlags,
					positions, orientations);

			framesList.add(skel);
		}
		currentAnimation.setFrames(framesList);

		return currentAnimation;
	}

	private MD5Frame buildSkeletonFrame(List<String> names, double[] frameData,
			int[][] jointFlags, List<Vector3> positions,
			List<Quaternion> orientations) {

		MD5Frame skeleton = new MD5Frame(jointFlags.length);

		for (int i = 0; i < jointFlags.length; i++) {
			Vector3 animatedPos = new Vector3(positions.get(i));
			Quaternion animatedOrient = new Quaternion(orientations.get(i));
			int j = 0;

			/* Tx */
			if ((jointFlags[i][1] & 1) == 1) {
				animatedPos.setI(frameData[jointFlags[i][2] + j]);
				++j;
			}

			/* Ty */
			if ((jointFlags[i][1] & 2) == 2) {
				animatedPos.setJ(frameData[jointFlags[i][2] + j]);
				++j;
			}

			/* Tz */
			if ((jointFlags[i][1] & 4) == 4) {
				animatedPos.setK(frameData[jointFlags[i][2] + j]);
				++j;
			}

			/* Qx */
			if ((jointFlags[i][1] & 8) == 8) {
				animatedOrient.setX(frameData[jointFlags[i][2] + j]);
				++j;
			}

			/* Qy */
			if ((jointFlags[i][1] & 16) == 16) {
				animatedOrient.setY(frameData[jointFlags[i][2] + j]);
				++j;

			}

			/* Qz */
			if ((jointFlags[i][1] & 32) == 32) {
				animatedOrient.setZ(frameData[jointFlags[i][2] + j]);
				++j;
			}

			computeW(animatedOrient);

			/*
			 * orientation NOTE: we assume that this joint's parent has already
			 * been calculated, i.e. joint's ID should never be smaller than its
			 * parent ID.
			 */

			MD5Joint parent = null;

			// for joints other than root joints figure out
			// what actual position and orientation is
			if (jointFlags[i][0] != -1) {
				parent = skeleton.getJoint(jointFlags[i][0]);

				Vector3 rpos = new Vector3();

				// add positions
				parent.getOrientation().rotatePoint(animatedPos, rpos);

				animatedPos.setI(rpos.getI() + parent.getPos().getI());
				animatedPos.setJ(rpos.getJ() + parent.getPos().getJ());
				animatedPos.setK(rpos.getK() + parent.getPos().getK());

				// concatenate rotations
				animatedOrient = parent.getOrientation().multiplication(
						animatedOrient);
				animatedOrient.normalize();
			}

			MD5Joint joint = new MD5Joint(names.get(i), parent, animatedPos,
					animatedOrient);

			skeleton.addJoint(i, joint);
		}

		return skeleton;
	}

	private static void computeW(Quaternion q) {
		double t = 1.0 - q.dotProduct(q);

		if (t < 0.0)
			q.setW(0.0);
		else
			q.setW(-Math.sqrt(t));
	}
}
